//
// Terminals.cpp
//

// Terminal controls that implement policy on top of the NVT.

#include "Terminals.h"
#include <coemain.h>
#include <coeaui.h>
#include <eikenv.h>
#include <eikdef.h>
#include <eikedwin.hrh>
#include <eikappui.h>
#include <eikapp.h>
#include <eikgted.h>
#include "jtelnet.h"
#include "NVT.h"

const TUid CTTY::KUidBufferLimit = { CJTelnetApp::BufferLimit };

//
// CTTY implementation
//


//
// Construction
//

CTTY* CTTY::NewLC(const TRect& aRect, MTerminalObserver* aObserver) {
    CTTY* self = new (ELeave) CTTY();
    CleanupStack::PushL(self);
    self->ConstructL(aRect, aObserver);

    return self;
}

CTTY::CTTY() 
: CCoeControl()
, MCoeControlObserver()
, MNVTObserver()
, MTerminal()
{
}


//
// Destructor
//
CTTY::~CTTY() {
    delete iEdwin;
    delete iNVT;
}


//
// ConstructL
//
// second-phase constructor stuff to initialize the model
void CTTY::ConstructL(const TRect& aRect, MTerminalObserver* aObserver) {

    iTermObserver = aObserver;

    iEdwin = new (ELeave) CEikGlobalTextEditor();
    CleanupStack::PushL(iEdwin);

    iNVT = CNVT::NewLC(this);

    CreateWindowL();

    iBufferLimit = STATIC_CAST(CJTelnetApp*, iEikonEnv->EikAppUi()->Application())->GetPrefTInt32L(KUidBufferLimit, 80 * 25 * 2);

    iEdwin->SetContainerWindowL(*this);
    iEdwin->SetObserver(this);
    iEdwin->ConstructL(this,0,iBufferLimit,EEikEdwinNoWrap | EEikEdwinInclusiveSizeFixed, EEikFontControlAll, EEikMonospaceFontsOnly);
    
    // set rectangle to prescription
    SetRectL(aRect);
    
    TCharFormat     charFormat(_L("MonoSpaced"), 150);
    TCharFormatMask charFormatMask;

    charFormatMask.SetAttrib(EAttFontTypeface);
    charFormatMask.SetAttrib(EAttFontHeight);
    iEdwin->ApplyCharFormatL(charFormat, charFormatMask);

    // Set line spacing to match text
    CParaFormat*    paraFormat = CParaFormat::NewLC();
    TParaFormatMask paraFormatMask;

    paraFormat->iLineSpacingInTwips = 150; 
    paraFormat->iLineSpacingControl = CParaFormat::ELineSpacingAtLeastInTwips;
    paraFormatMask.SetAttrib(EAttLineSpacing);
    paraFormatMask.SetAttrib(EAttLineSpacingControl);
    iEdwin->ApplyParaFormatL(paraFormat, paraFormatMask);
   
    CleanupStack::PopAndDestroy();  // paraFormat

    iPhase = ETermDisconnected;

    ActivateL();
    iEdwin->SetFocus(ETrue);
    CleanupStack::Pop(2);
}


CCoeControl* CTTY::ComponentControl(TInt aIndex) const { 
    return (aIndex < CountComponentControls() ? iEdwin : NULL); 
}



//
// HandleControlEventL
//
// accept and handle focus resquests from our components.
void CTTY::HandleControlEventL(CCoeControl* aControl,TCoeEvent aEventType) {
    if (aEventType==EEventRequestFocus) {
	aControl->SetFocus(ETrue);
    }
}


TKeyResponse CTTY::OfferKeyEventL( const TKeyEvent& aKeyEvent
				 , TEventCode       aType
				 ) 
{

    if ((aKeyEvent.iModifiers & (EModifierFunc | EModifierCtrl)) == (EModifierFunc | EModifierCtrl))
    {
	if (aKeyEvent.iCode && aType == EEventKey) 
	{
	    iNVT->WriteL(aKeyEvent.iCode);
	}
    }
    else 
    {
	if (aType!=EEventKey) 
	{
	    return EKeyWasNotConsumed;
	}

	iNVT->WriteL(aKeyEvent.iCode);
    }
    return EKeyWasConsumed;
}

void CTTY::ActivateL() {
    CCoeControl::ActivateL();

    iEdwin->ActivateL();
}

// only leave on error - use result to determine network operation success
void CTTY::ConnectL(const TDesC& aAddr, const TUint aPort) {

    PhaseChange(iPhase, ETermConnecting);
    iNVT->ConnectL(aAddr, aPort);
}


void CTTY::DisconnectL(TDisconnectType aType) {
    
    PhaseChange(iPhase, ETermDisconnecting);

    CNVT::TDisconnectType type = aType == EDisconnectNormal ? CNVT::EDisconnectNormal : CNVT::EDisconnectNet;
    iNVT->DisconnectL(type);
}

void CTTY::PhaseChange(EConnectPhase prev, EConnectPhase current) {
    if (iTermObserver) {
	iTermObserver->NotifyPhaseChange(prev, current);
    }

    iPhase = current;
}


MTerminal::EConnectPhase CTTY::Phase() const {
    return iPhase;
}



void CTTY::SizeChangedL() {
    iEdwin->SetRectL(Rect());
}


TInt FindFirstOf(const TDesC& aTarget, const TDesC& aTokens) {

    if (aTokens.Length() < 1) {
	return KErrNotFound;
    }
    else if (aTokens.Length() == 1) {
	return aTarget.Locate(aTokens[0]);
    }
    else {
	for (TInt i = 0; i < aTarget.Length(); ++i) {
	    if (aTokens.Locate(aTarget[i]) >= 0) {
		return i;
	    }
	}
	return KErrNotFound;
    }
}

void CTTY::OnBufferUpdateL(const TDesC& aBuf) {

    iEdwin->TextView()->SetPendingSelection(TCursorSelection(iEdwin->Text()->DocumentLength(),iEdwin->Text()->DocumentLength()));
    const TInt cursorPos = iEdwin->CursorPos();

    HBufC* insterted_chars = HBufC::NewLC(aBuf.Length() * 2);	// worst case we double it.

    for (TInt i = 0; i < aBuf.Length(); ++i) {

	if (aBuf[i] > 31 && aBuf[i] < 127) {	// printable ascii
	    insterted_chars->Des().Append(aBuf[i]);
	}
	else {
	    if (aBuf[i] == 0x0D) {
		insterted_chars->Des().Append(CPlainText::ELineBreak);
	    }
	    else if (aBuf[i] == 0x0A) { // skip it.
	    }
	    else if (aBuf[i] == 0x08) {
		// backspace
		if (i == 0) {
		    TInt length = iEdwin->Text()->DocumentLength();
		    iEdwin->Text()->DeleteL(length - 1, 1);
		    iEdwin->TextView()->HandleCharEditL(CTextLayout::EFLeftDelete);
		}
		else {
		    insterted_chars->Des().Delete(insterted_chars->Des().Length() - 1, insterted_chars->Des().Length());
		}
	    }
	    else if (aBuf[i] > 0 && aBuf[i] < 27) {
		insterted_chars->Des().Append('^');
		insterted_chars->Des().Append(aBuf[i] + 64);
	    }
	    else {
		// silently drop anything else - we're a plain-old-tty
	    }
	}
    }

    if (iEdwin->Text()->DocumentLength() + insterted_chars->Des().Length() > iBufferLimit) {

	iEdwin->Text()->DeleteL(0, insterted_chars->Des().Length());
    }
    iEdwin->Text()->InsertL(iEdwin->Text()->DocumentLength(), insterted_chars->Des());

    const TInt newCursorPos = iEdwin->Text()->DocumentLength();
    iEdwin->TextView()->SetPendingSelection(TCursorSelection(newCursorPos,newCursorPos));
    iEdwin->TextView()->HandleInsertDeleteL(TCursorSelection(newCursorPos,cursorPos),0,ETrue);

    // make sure the cursor is at the base of a full screen.
    TInt YPos = Rect().Height();
    TViewYPosQualifier qual;
    qual.SetMakeLineFullyVisible();
    iEdwin->TextView()->SetViewL(iEdwin->TextLength(), YPos, qual);

    CleanupStack::PopAndDestroy();
}

void CTTY::OnBufferUpdateL(const TChar aChar) {

    TBuf<1> temp(1);
    temp[0] = TUint8(aChar);

    OnBufferUpdateL(temp);
}


// return ETrue to proceed to the next phase
void CTTY::OnServerStart(TRequestStatus aStatus) {
    if (iTermObserver) {
	iTermObserver->OnServerStart(aStatus);
    }

    if (aStatus != KErrNone) {
        PhaseChange(iPhase, ETermDisconnected);
    }
}

void CTTY::OnAddressResolved(TRequestStatus aStatus) {
    if (iTermObserver) {
	iTermObserver->OnAddressResolved(aStatus);
    }

    if (aStatus != KErrNone) {
        PhaseChange(iPhase, ETermDisconnected);
    }
}

void CTTY::OnConnect(TRequestStatus aStatus) {

    if (iTermObserver) {
	iTermObserver->OnConnect(aStatus);
    }

    if (aStatus == KErrNone) {
        PhaseChange(iPhase, ETermConnected);
    }
    else {
        PhaseChange(iPhase, ETermDisconnected);
    }
}


void CTTY::OnDisconnect(TRequestStatus aStatus) {
    if (iTermObserver) {
	iTermObserver->OnDisconnect(aStatus);
    }

    PhaseChange(iPhase, ETermDisconnected);
}


void CTTY::FocusChanged(TDrawNow /* aDrawNow */) {

}

void CTTY::ClipboardL(TClipboardFunc aClipboardFunc) {
    switch (aClipboardFunc) {
    case ECopy:
	iEdwin->ClipboardL(CEikEdwin::ECopy);
	break;

    case EPaste:
	iEdwin->ClipboardL(CEikEdwin::EPaste);
	break;
    }
}


void CTTY::SelectAllL() {
    iEdwin->SelectAllL();
}


void CTTY::SetZoomFactorL(TZoomFactor* aZF) {
    iEdwin->SetZoomFactorL(aZF);
}


void CTTY::SetZoomLevelL(TZoomLevel) {

    iEdwin->NotifyNewFormatL();
    iEdwin->DrawNow();
}
